//---------------------------------------------------------------------------------------
// Контроллер куба
// v.05.10.2016
//---------------------------------------------------------------------------------------
#include <Wire.h>

#define SLAVE_ADDRESS       0x04        // Адрес шины i2c для связи с внешним миром. Slave.

// Константы - режимы работы контроллера. 
// Режим выбирается автоматически в зависимости от наличия детектора нуля сетевого напряжения.
// Сигнал детектора нуля ожидается на пине D2/INT0
#define MODE_PWM            0b00000000  // ШИМ-регулировка мощности с периодом 1 сек. Этот режим включается если 
                                        // детектор нуля не подключен к пину D2/INT0
#define MODE_PDM            0b00000001  // Pulse Density Modulation (PDM) - регулировка мощности за счет
                                        // изменения плотности импульсов. Метод диффузии ошибки (ака - метод Брезенхема)
                                        // Для этого режима необходимо наличие сигнала детектора нуля на входе D2/INT0
// Константы - состояние кнопок
#define NO_ACT              0
#define PLUS_CLICK          0b00000001  // Короткое нажатие кнопки "+"
#define MINUS_CLICK         0b00000010  // Короткое нажатие кнопки "-"
#define PLUS_PRESS          0b00000100  // Удержание кнопки "+" в нажатом состоянии
#define MINUS_PRESS         0b00001000  // Удержание кнопки "-" в нажатом состоянии
// Номера пинов
static uint8_t mButPin = 0,             // Пин, к которому подключена кнопка "-"
               pButPin = 1,             // Пин, к которому подключена кнопка "+"
               outPin = 11,             // Пин выходного сигнала (для управления симистором)
               inPin = 2;               // Пин сигнала от детектора нуля (используется внешнее прерывание INT0)
// Пины, к которым подключены сегменты индикатора (аноды)
static uint8_t segPins[7] = {13, 17, 4, 6, 7, 14, 3};
// Пины, к которым подключены общие катоды цифр индикатора
static uint8_t digPins[3] = {16, 15, 12};
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static int d[3]; // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int curSeg = 0; // Сегмент, обрабатываемый (подствечиваемый, если нужно) в данный момент
static int prSeg = 6; // Предыдущий сегмент
static int dt = 1; // Задержка (мс) при выводе сегментов индикатора

static int dtBounce = 10; // Задержка (мс) для устранения дребезга контактов кнопок
static int dtPress = 300; // Задержка, после которой считается, что кнопка удерживается
static uint8_t bos = NO_ACT; // Предыдущее состояние кнопок

static uint8_t power = 0; // Задаваемый уровень мощности в %
static uint8_t powerChangedFlag = 1; // Индикатор того, что значение мощности было изменено при помощи кнопок или i2c 
static uint8_t dPower = 5;  // Приращение устанавливаемой кнопками мощности при удержании кнопок нажатыми
static unsigned long tpr; // Момент нажатия кнопки

static uint8_t pFlag = 0; // Наличие сигнала (импульса) на выходе модулятора
static unsigned long pStart; // Момент начала импульса на выходе модулятора
static unsigned long pWidth; // Длительность управляющего импульса в мс. Для режима PDM это будет всегда 5 мс,
                             // а в режиме ШИМ - в зависимости от значения power.
static unsigned long T = 1000; // Период ШИМ-сигнала (для режима MODE_PWM). В данном случае - 1 сек.
static int lev = 0, err = 0; // Вспомогательные переменные для реализации PDM методом диффузии (смещения) ошибки

static uint8_t mode = MODE_PWM; // Режим работы контроллера. ШИМ по умолчанию.
static uint8_t firstInt = 1; // Флаг для отслеживания самого первого аппаратного прерывания

//---------------------------------------------------------------------------------------
void Int0() { // Функция - обработчик внешнего прерывания INT0
  if(firstInt) { // Уж коли сюда попали - значит детектор нуля подключен. Следовательно будем работать в режиме PDM
    firstInt = 0; mode = MODE_PDM; pWidth = 5; // В режиме PDM длительность управляющего импульса всегда будет 5 мс
  }

  // Небольшая задержка 100 мкс. Ну просто для включения затвора семистора поближе к нулю.
  // Эту задержку желательно подобрать в зависимости от параметров конкретного детектора нуля)
  unsigned long st = micros() + 100;
  while(micros() < st); // В данном случае ждем 100 мкс

  lev = power + err; // Текущий уровень с учетом ошибки дискретизации, сделанной на предыдущем полупериоде.
  if(lev >= 50) {
    digitalWrite(outPin, HIGH); err = lev - 100;  // Считаем ошибку для следующего полупериода
  } else {
    digitalWrite(outPin, LOW); err = lev;  
  }
//  lev = power + err;
  pStart = millis(); pFlag = 1; // Фиксируем момент начала управляющего импульса (или его отсутствия)
}
//---------------------------------------------------------------------------------------
void setup() {
  pinMode(mButPin, INPUT); // Инициализация пинов, к которым подключены кнопки
  digitalWrite(mButPin, HIGH);
  pinMode(pButPin, INPUT);
  digitalWrite(pButPin, HIGH);

  pinMode(outPin, OUTPUT); // Иницализация пина выходного (модулированного) сигнала
  digitalWrite(outPin, LOW);

  pinMode(inPin, INPUT); // Инициализация пина для входного сигнала
  digitalWrite(inPin, HIGH);
  
  attachInterrupt(0, Int0, FALLING); // Прерывания по спаду импульсов входного сигнала
//  attachInterrupt(0, Int0, RISING); // Прерывания по фронту импульса входного сигнала

  Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
  Wire.onReceive(receiveData);
  Wire.onRequest(sendData);

  for(int i = 0; i < 7; i++) { // Инициализация пинов, к которым подключены аноды индикатора (сегменты)
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], LOW);
  }
  
  for(int i = 0; i < 3; i++) { // Инициализация пинов, к которым подключены катоды индикатора (цифры)
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], HIGH);
  }

  unsigned long st = millis() + 1000;
  d[0] = 8; d[1] = 8; d[2] = 8; // Для демонстрации работоспособности индикатора покажем все сегменты разом
  while(millis() < st) {
      ShowSegments();
  }
  d[0] = 10; d[1] = 10; d[2] = 10; // Очистим индикатор
  st = millis() + 500;
  while(millis() < st) {
      ShowSegments();
  }
  // К этому моменту уже должно пройти (или не пройти) первое прерывание для определения 
  // режима работы контроллера. Т.е. режим работы уже выбран. Покажем же выбор системы на индикаторе...
  // Мигающий 0 - это PWM, мигающая 1 - это PDM
  for(int i = 0; i < 5; i++) {
    st = millis() + 500;
    d[0] = (!(i % 2))? mode : 10;
    while(millis() < st) {
        ShowSegments();
    }
  }
  d[0] = 10; // Опять очистим индикатор
  st = millis() + 500;
  while(millis() < st) {
      ShowSegments();
  }

  // Заканчиваем инициализацию и переходим к штатному режиму работы системы
  InitDigits(power);
}
//---------------------------------------------------------------------------------------
void loop() {
  // Анализируем состояние кнопок
  uint8_t ks = GetKey();
  switch(ks) {
    case PLUS_CLICK:  if(power >= 100) power = 100;
                      else power++;
                      powerChangedFlag = 1;
                      break;
    case MINUS_CLICK: if(power == 0) power = 0;
                      else power--;
                      powerChangedFlag = 1;
                      break;
    case PLUS_PRESS:  if(power >= 100 - dPower) power = 100;
                      else power += dPower;
                      powerChangedFlag = 1;
                      break;
    case MINUS_PRESS: if(power <= dPower) power = 0;
                      else power -= dPower;
                      powerChangedFlag = 1;
                      break;
  }

  if(powerChangedFlag) // Скорректируем показания индикатора, если мощность была изменена
    InitDigits(power);
  
  ShowSegments(); // Постоянно занимаемся подсвечиванием сегментов индикатора

  // И следим за уровнем управляющего сигнала на выходе
  unsigned long now = millis();
  // Если время управляющего импульса истекло - выключим его
  if(pFlag && now - pStart > pWidth) {
    digitalWrite(outPin, LOW);
    pFlag = 0;
  }
  // Следующий фрагмент - только для режима PWM. Если закончился период ШИМ (1 сек) - включаем 
  // новый импульс (а в режиме PDM импульс включается в функции обработки внешних прерываний Int0())
  if(mode == MODE_PWM && now - pStart >= T) {
    if(power > 0)
      digitalWrite(outPin, HIGH);
    pFlag = 1; pStart = now; pWidth = 10*power;
  }
}
//---------------------------------------------------------------------------------------
uint8_t GetKey(void) {
  uint8_t ret, bcs = ((!digitalRead(mButPin) & 0x01) << 1) | (!digitalRead(pButPin) & 0x01); // Текущее состояние кнопок
  
  if(!bos && bcs) { // Произошло нажатие кнопки
    tpr = millis(); bos = bcs; // Фиксируем время и состояние кнопок
    return NO_ACT;
  }
  if(bos && !bcs) { // Произошло либо отжатие или дребезг кнопки
    if((millis() - tpr) < dtBounce) return NO_ACT; // Дребезг, однако...
    ret = bos; bos = bcs; return ret;
  }
  if(bos && bcs) { // Возможно, удерживание кнопки
    if((millis() - tpr) < dtPress) return NO_ACT; // Но пока еще не ясно...
    // И все-таки это удержание кнопки в нажатом состоянии
    ret = bos << 2; tpr = millis(); return ret;
  }
  return NO_ACT;
}
//---------------------------------------------------------------------------------------
void InitDigits(int val) { // Подготовка данных (цифр) для визуализации
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) { // Не отображаем ведущие нули ежели что... ("Цифра" 10 - это "пустая" цифра, все ее сегменты потушены)
    d[2] = 10;
    if(d[1] == 0) d[1] = 10;
  }
  powerChangedFlag = 0;
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Вывод текущего сегмента на разрядах 3-х разрядный 7-сегментный индикатор
  // Посегментное динамическое подсвечивание светодиодов

  digitalWrite(segPins[curSeg], HIGH); // "Подбрасываем" потенциал аноду текущего сегмента
  digitalWrite(segPins[prSeg], LOW); // "Роняем" потенциал анода предыдущего сегмента

  for(int bi = 0; bi < 3; bi++) { // Теперь пробегаем по цифрам (разрядам) индикатора
    if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], LOW); // то уроним потенциал катода соответствующей цифры (разряда)
    else // А если гореть не должен, то
      digitalWrite(digPins[bi], HIGH); // потенциал катода bi-й цифры должен быть высокий 
  }
  delay(dt); // Посветим немножко тем, что должно светиться...
  mask <<= 1; // Корректируем маску для следующего сегмента
  prSeg = curSeg;
  if(++curSeg > 6) {
    curSeg = 0; prSeg = 6; mask = 0x01;
  }
}
//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов
  while(Wire.available()) {
    power = Wire.read(); powerChangedFlag = 1;
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
    Wire.write(power);
}
//---------------------------------------------------------------------------------------

